/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

///////////////
// murasaki - multiple genome global alignment program
// by Kris Popendorf (krisp@dna.bio.keio.ac.jp)
/////////////
// murasaki.cc
// main program control and such
//////////

#include <iostream>
#include <fstream>
#include <vector>
#include <assert.h>
#include <iterator>
#include <unistd.h>
#include <limits.h>
#include <math.h>
#include <pstream.h>
#include "murasaki.h"
#include "timing.h"
#include "dinkymath.h"
#include "sequence.h"
#include "msethash.h"
#include "dbhash.h"

#include <boost/regex.hpp> //screw "sscanf" =P
#include <boost/lexical_cast.hpp>

#ifdef __FreeBSD__
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/sysctl.h>
#include <sys/vmmeter.h>
#endif

#ifdef __APPLE__
#include <mach/mach_init.h>
#include <mach/host_info.h>
#include <mach/mach_host.h>
#endif

#ifdef MURASAKI_MPI
#include "mpi.h"
#endif

using namespace std;
int seq_count=0;
Hash *mfh; //multi-fasta-hash
vector<Sequence*> seqs;
word usedBuckets=0,usedPats=0,worstPat=0;
UsedMap *anchors;
string version_string(PROGRAMVERSION);
string pattern_src("111010010100110111");//the pattern hunter 18long
Options opt; //runtime option manager

int debugbits=-22;
SystemInfo sysinfo;
ProgressTicker ticker(100);

int verbose=0; //ubersneaky debug switch

#ifdef MURASAKI_MPI
int mpi_id;
int mpi_procs;
bool mpi_capable=false;
#endif

int main (int argc,char **argv){

#ifdef MURASAKI_MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_id);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_procs);

  // display hostname
  char nodename[81];
  gethostname(nodename, 80);

  cout << "murasaki/MPI process #" << mpi_id << ": running on " << nodename << endl;

  // redirection
  int stdoe=0; // stdout & err
  char stdoename[101];

  if(mpi_id != 0){ // only process #0 uses the console
    snprintf(stdoename, 100, "murasaki-mpiout-%d", mpi_id);
    stdoe = open(stdoename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
    dup2(stdoe, 1);
    dup2(stdoe, 2);
  }
#endif

  initConstants();

  //option procesing
  opt.commandline_opts(argc,argv);
  //grab remaining
  vector<string> args;

  if(optind<argc){ //if custom args set
    cout << "Custom args: ";
    for(int i=optind;i<argc;i++){
      cout << argv[i] << " ";
      args.push_back(string(argv[i]));
    }
    cout << endl;
  }else{
#ifdef NDEBUG
    //a full release would do this...
    //    cout << program_help(); exit(1);
    //but...we're not a "full" release yet....
    args.push_back(string("seq/CDC1551.gbk"));
    args.push_back(string("seq/leprae.gbk"));
    pattern_src="11111100100001111100000100101111";
    cout << "Default test (release) run: "<<args[0]<<" vs "<<args[1] << endl;
#else
 //run default tests
    args.push_back(string("seq/test1.gbk"));
    args.push_back(string("seq/test2.gbk"));
    cout << "Default test run: "<<args[0]<<" vs "<<args[1] << endl;
#endif
  }

  //generate pattern from pattern_src if necessary
  convertPattern(pattern_src);
  writeOut(opt.pat_record,pattern_src);

  //solidfy parameters
  seq_count=args.size();
  opt.solidify();
  ofstream opf(opt.options_record.c_str());
  opf << opt;
  opf << platformInfo();
  opf.close();
  cout << opt;

#ifdef MURASAKI_MPI
  // synchronize random seed
  if(opt.hashPref == ArrayHash_t){
    mpi_capable = true;
    MPI_Bcast((void*)&opt.rand_seed, sizeof(int), MPI_CHAR, 0, MPI_COMM_WORLD);
  }
#endif
  
  writeOut(opt.status_record,"Loading sequences...");
  int longestSeq=0;
  int totalSeqLength=0;
  int totalHashLength=0; //include both fwd/rev for all but seq[0]
  //load sequences
  ofstream outf(opt.seq_record.c_str());
  Timer seqLoadStart;
  for(int i=0;i<seq_count;i++){
    outf << args[i] << endl;
    seqs.push_back(new Sequence(args[i]));
    cout << "Seq "<<seqs[i]->name<<" loaded. "<<humanMemory(seqs[i]->length())<<endl;
    longestSeq=max<int>(longestSeq,seqs[i]->length());
    totalSeqLength+=seqs[i]->length();
    totalHashLength+=(i==0 ? seqs[i]->length():seqs[i]->length()*2);
  }
  set_seqtable(seqs);  // prepare for hashing: can we skip this process in future version?
  Timer seqLoadStop;
  outf.close();
  cout << "Sequence loading took: "<<elapsed(seqLoadStart,seqLoadStop)<<endl;

  //ready pattern(s) and hash
  BitSequence *pat=new BitSequence(pattern_src);
  cout << "Pat: ("<<pat->length()<<" bases long) contains "<<pat->hashLength()<<" bits"<<endl;
  cout << *pat <<endl;
  int hashLength=pat->hashLength();
  double idealHash=pow((double)4.,(double)(hashLength));
  //  hash_size=idealHash > (double)hash_max ? hash_max:(int)idealHash;
  hash_max=intpow(2,max_hash_bits);
  int hashBits=min<int>(max_hash_bits,hashLength*2);

#ifdef MURASAKI_MPI
  // synchronize hashBits
  if(mpi_capable){
    if(mpi_id != 0){
      // send hashBits
      MPI_Send(&hashBits, sizeof(int), MPI_CHAR, 0, 0, MPI_COMM_WORLD);
    } else {
      // collect hashBits
      for(int i=1; i<mpi_procs; i++){
	int hb;
	MPI_Status status;
	MPI_Recv(&hb, sizeof(int), MPI_CHAR, i, 0, MPI_COMM_WORLD, &status);
	hashBits=min<int>(hashBits, hb);
      }
    }
    // synchronize
    MPI_Bcast(&hashBits, sizeof(int), MPI_CHAR, 0, MPI_COMM_WORLD);
  }
#endif

  Timer hashFuncCreate;
  writeOut(opt.status_record,"Computing hash parameters...\n");
  cout << "Computing hash parameters...\n";
  initHashParams(hashBits,pat);
  writeOut(opt.status_record,"Creating storage structures...\n");
  Timer hashCreate;
  cout << "Hash parameters computed in "<<elapsed(hashFuncCreate,hashCreate)<<endl;
  switch(opt.hashPref){
  case ArrayHash_t: mfh=new ArrayHash(pat);break;
  case MSetHash_t: mfh=new MSetHash(pat);break;
#ifdef MURASAKI_MYSQL
  case MySQLHash_t: mfh=new MySQLHash(pat); break;
#endif
#ifdef MURASAKI_SQLITE3
  case SQLiteHash_t: mfh=new SQLiteHash(pat); break;
#endif
  default: cout << "That hash ("<<opt.hashPref_s()<<") isn't doesn't exist."<<endl
		<< "Using ArrayHash instead."<<endl;
    mfh=new ArrayHash(pat);break;
  }
  anchors=new UsedMap(seq_count);
  Timer hashStart;
  cout << "Storage structures created in: "<<elapsed(hashCreate,hashStart)<<endl;


  //report size info after the fact...
  cout << "Hash blocks are: "<<humanMemory(mfh->linear_cost())<<" + 1-time cost of "<<humanMemory(mfh->new_bucket_cost())<<endl;
  unsigned long linearHashCost=mfh->linear_cost()*totalHashLength;
  unsigned long linearAnchorCost=sizeof(multiset<Space,ltSpace>)*totalSeqLength+sizeof(vector<multiset<Space,ltSpace> >)*seq_count;
  unsigned long hashBlockSize=mfh->bucket_prep_cost();
  unsigned long bucketPrepCost=hashBlockSize*hash_size;
  unsigned long avgCaseMemory=bucketPrepCost+mfh->new_bucket_cost()*hash_size/2+linearHashCost/2+linearAnchorCost/2;
  unsigned long bestCaseMemory=bucketPrepCost+sizeof(multiset<HashVal>)*seq_count;
  unsigned long committedMemory=bucketPrepCost+linearHashCost+linearAnchorCost;

  cout << "Would take minimum "<< humanMemory(idealHash*hashBlockSize) <<" to make perfect hash.\n"
       << " + linear (in seq length) hashing cost of "<<humanMemory(linearHashCost)<<endl
       << " + linear (in seq length) expected anchor cost of "<<humanMemory(linearAnchorCost)<<endl 
       << "We could do "<< humanMemory((unsigned long)hash_max*hashBlockSize+linearHashCost+linearAnchorCost) <<" so, using: " << humanMemory(committedMemory) << "("<<hashBits<<" bits, "<<hash_size<<" entries)"<< endl;
  cout << "Expected memory usage: "<<humanMemory(avgCaseMemory)<<endl;
  if(committedMemory/1024>sysinfo.freeMemory)
    cout << "**** Warning: This is more memory you have free on your system.\n"
	 << "****          You may encounter massive slowdowns if murasaki\n"
	 << "****          gets swapped to disk.\n"
	 << "****--------> Consider changing hash settings.\n";
  if(avgCaseMemory/1024>sysinfo.totalMemory)
    cout << "****>>>>>> Impending Disaster <<<<<<****" << endl
	 << "****>> Murasaki "<<
      (bestCaseMemory/1024>sysinfo.totalMemory ? "_will_":"may")
	 <<" consume more memory than you have in your system.\n"
	 << "****>> Consider changing hash settings.\n";

  //archive region list
  ofstream regionsOf(opt.region_record.c_str());
  outputRegions(regionsOf);
  regionsOf.close();
  
  //hash the monkey! all but the first of the monkies...
  for(unsigned i=(opt.doubleHash && opt.histogram ? 0:1);i<seqs.size();i++){
#ifdef MURASAKI_MPI
    // skip if the sequence isn't mine
    if( mpi_capable && (int)i%mpi_procs != mpi_id ) continue;
#endif

    // try hashCache
    if(opt.hashCache){
      mfh->clear();
      string cachename=hashCacheName(seqs[i]->name, pat); 
      ifstream cachef;
      cachef.open(cachename.c_str(), ios::binary);
      if (cachef){ 
	cachef.close();
	cout << "cache hit, so skipping to hash " << seqs[i]->name << "." << endl; 
	continue; 
      }
    }

    // Hash 
    if(!opt.skipFwd){
      cout << "\nHashing "+seqs[i]->name+" forwards\n";
      writeOut(opt.status_record,"Hashing "+seqs[i]->name+" forwards\n");
      hashSeq(seqs[i]->fwd,pat,1,seqs[i]);
    }
    if(!opt.skipRev){
      cout << "\nHashing "+seqs[i]->name+" backwards\n";
      writeOut(opt.status_record,"Hashing "+seqs[i]->name+" backwards\n");
      hashSeq(seqs[i]->rev,pat,-1,seqs[i]);
    }

    // Save hashCache
    if(opt.hashCache){
      string cachename=hashCacheName(seqs[i]->name, pat); 
      ofstream cachef; 
      cout << endl <<"Saving hash as: " << cachename << endl;
      cachef.open(cachename.c_str(), ios::trunc|ios::out|ios::binary);
      if(!cachef){ cerr << "can't write " << cachename << endl; exit(-1); }
      mfh->dump(cachef);
      cachef.close();
    }
    //    assert(mfh->sanityCheck());
  }

  // load from cache
  if(opt.hashCache){
    mfh->clear();
    for(unsigned i=(opt.doubleHash && opt.histogram ? 0:1);i<seqs.size();i++){
      string cachename=hashCacheName(seqs[i]->name, pat);
      ifstream cachef; 
      cout << "Loading hash from: " << cachename << endl;
      cachef.open(cachename.c_str(), ios::binary);
      if(!cachef){ cerr << "can't load " << cachename << endl; exit(-1); }
      mfh->load(cachef);
      cachef.close();
    }
  }

  Timer hashStop;
  cout << "\nHashing all but first sequence took: "<<elapsed(hashStart,hashStop)<<endl;

#ifdef MURASAKI_MPI
  MPI_Barrier(MPI_COMM_WORLD);
  if (mpi_capable && mpi_id != 0){
    writeOut(opt.status_record,"MPI client mode: accepting queries...\n");
    mpi_client_mode(); 
    close(stdoe);
    writeOut(opt.status_record,"MPI client mode: terminated.\n");
    MPI_Finalize();
    exit(0);
 }
#endif

  word &hash=activeHash;
  word hits=0,multi=0;
  int lastStop=seqs[0]->fwd->matchRegions.back().second-pat->length()+1;
  writeOut(opt.status_record,"Hashing "+(seqs[0]->name)+" for real.\n");
  cout << "Hashing "+seqs[0]->name+" forwards and computing anchors...\n";
  ticker.reset(lastStop);
  for(vector<pair<int,int> >::iterator region=seqs[0]->fwd->matchRegions.begin();
      region!=seqs[0]->fwd->matchRegions.end();region++){
    int start=region->first+1;
    int stop=region->second-pat->length()+2;
    Window fwd(seqs[0]->fwd,region->first,pat); //for last strand only do fwd
    word fullMclass=lowN(seq_count);
    for(int p=start;p<=stop;p++){
#ifdef NDEBUG
      int progress=ticker.tick(p-1);
      if(progress>=0){
	hashStatusCheck(progress,p);
      }
#else
      lastStop=lastStop; //stupid "no effect" warnings...
#endif

      assert(cout << "Hashing "<<p);
      hash=fwd.hash();
      assert(cout << " - " << hash << " = "<<wordToString(hash)<<endl);
      //      assert(cout << "Hashing "<<seqs[0]->name<<"."<<"fwd"<<":"<<p<<"->"<<wordToString(hash)<<endl);
      HashVal here(seqs[0],p);
      int size=mfh->sizeAt(hash);
      if(size){
	if(opt.hashfilter){
	  if(size>=opt.hashfilter)
	    goto NextHash;
	}
	if(opt.seedfilter){
	  if(mfh->sizeAt(hash,here)>=(unsigned)opt.seedfilter)
	    goto NextHash;
	}
      }
      if(opt.histogram && !opt.doubleHash){
	mfh->add(hash,here);
	//    assert(mfh->sanityCheck());
      }
      size++;
      //      assert(mfh->sanityCheck(hash));
      if(size>=seq_count && !opt.hashOnly){ //potential hit
	LocList locList(seq_count);
	mfh->lookup(hash,here,locList);
	//this is probably faster than leaving it up to the full n^s search
	//(though in preliminary testing it showed virtually nochange)
	if(locList2mclass(locList)==fullMclass){
	  list<Location> start;
	  //tweakup loclist (we only need to process the most recent hit on seq[0])
	  locList[0].clear();
	  locList[0].push_back(here);
	  procLocs(locList,start);
	  //      assert(anchors->sanityCheck());
	}
      }
      //    assert(anchors->sanityCheck());
    NextHash:
      multi+=size;
      hits++;
      fwd.slide();
    }
  }
  cout << endl;
  if(opt.histogram>1 && !opt.skipRev && !opt.doubleHash){ //need to also hash seq[0] backwards to make histogram stats correct
    cout << "Hashing "+seqs[0]->name+" backwards for statistical completeness...\n";
    hashSeq(seqs[0]->rev,pat,-1,seqs[0]);
    cout << endl;
  }
  cout << "Total anchors: "<<anchors->count()<<endl;
  cout << "Average bucket depth: " <<fstring((double)multi/(double)hits)<<endl;
  Timer matchStop;
  cout << "Last sequence hash + match: "<<elapsed(hashStop,matchStop)<<endl;
  cout << "Total processing time: "<<elapsed(seqLoadStart,matchStop)<<endl;
  Timer outputWrite;
  if(opt.histogram){
    cout << "Writing histogram data...";cout.flush();
    writeHisto(opt.hashHisto_record,opt.hashDetailed_record);
  }else{
    cout << "Counting buckets...";cout.flush();
    mfh->writeHistogram(cerr); //phony write
  }
  cout << endl;
  if(!opt.hashOnly){
    cout << "Writing anchors...";cout.flush();
    writeAnchors(opt.anchor_record);
  }
  if(opt.histogram){
    double optimal=(double)totalSeqLength*2/(double)hash_size*100.0;
    double usage=(double)usedBuckets/(double)hash_size*100.0;
    cout << "Buckets used: "<<usedBuckets<<" ("<<fstring(usage)<<"% = "
	 <<fstring(usage/optimal*100.)<<"% of optimal "
	 <<fstring(optimal)<<"%)"<<endl;
  }
  if(opt.histogram>3){
    cout << "Total pattern encountered: "<<usedPats<<endl;
    double avgPats=(double)usedPats/(double)usedBuckets*100.0;
    cout << "Average patterns per hash: "<<avgPats<<endl;
  }


  Timer totalDone;
  cout << "Output writing finished in: "<<elapsed(matchStop,totalDone)<<endl;
  cout << "Total time: "<<elapsed(seqLoadStart,totalDone)<<endl;
  writeOut(opt.status_record,"Done!\n");

#ifdef MURASAKI_MPI
  mpi_req request;
  request.type = MURASAKI_MPI_REQ_QUIT;
  
  MPI_Bcast(&request, sizeof(request), MPI_CHAR, 0, MPI_COMM_WORLD);
  MPI_Finalize();
#endif
}

void writeAnchors(string anchor_record){
  ofstream of(anchor_record.c_str());
  of << *anchors;
  of.close();
  ofstream ancDetails((anchor_record+".details").c_str());
  if(opt.bitscore){
    ofstream ancBitscore((anchor_record+".bitscore").c_str());
    anchors->saveDetails(ancDetails,ancBitscore);
    ancBitscore.close();
  }else{
    anchors->saveDetails(ancDetails,cerr);
  }
  ancDetails.close();
}

void writeHisto(string hashHisto_record,string details){
  ofstream histo(hashHisto_record.c_str());
  if(opt.histogram)
    mfh->writeHistogram(histo);
  histo.close();
  if(opt.histogram>=3){
    cout << "Writing detailed histogram...";cout.flush();
    ofstream histo(details.c_str());
    mfh->writeDetailedHistogram(histo);
    histo.close();
  }
}

void hashSeq(BitSequence *bitseq,BitSequence *pat,int sign,Sequence *s){
  ticker.reset(bitseq->length()-1);
  for(vector<pair<int,int> >::iterator region=bitseq->matchRegions.begin();
      region!=bitseq->matchRegions.end();region++){
    //cout << "Valid region is "<<region->first<<" to "<<region->second<< " - patlen" << pat->length() << endl;
    int stop=region->second-pat->length()+2;
    int start=region->first+1;
    Window win(bitseq,region->first,pat);
    word hash=win.hash();
    int progress;
    // cout << "Hasing region from "<<start<<" to "<<stop<<endl;
    for(int p=start;p<=stop;p++){
      hash=win.hash();
      // assert(cout << "Hashing "<<s->name<<"."<<(sign>0 ? "fwd":"rev")<<":"<<p<<"->"<<wordToString(hash)<<endl);
      assert(cout << "At "<<p<<" "<<hash<<" hash= " << wordToString(hash)<<endl);
      Location here(s,p*sign);
      if(!mfh->emptyAt(hash)){
	   if(opt.hashfilter){
	     if(mfh->sizeAt(hash)>(unsigned)opt.hashfilter)
	       goto NextHash;
	   }
	   if(opt.seedfilter){
	     if(mfh->sizeAt(hash,here)>(unsigned)opt.seedfilter)
	       goto NextHash;
	   }
      }
      mfh->add(hash,here);
    NextHash:
      if((progress=ticker.tick(p-1))>=0){
	hashStatusCheck(progress,p);
      }
      // assert(mfh->sanityCheck(hash));
      win.slide();
    }
  }
}

word locList2mclass(LocList &locList){
  word mclass=0;
  int size=locList.size();
  for(int i=0;i<size;i++)
    if(!locList.empty())
      mclass|=1<<i;
  return mclass;
}

bool sanityCheck(Location &l){
  if(!l.seq())return false;
  if(l.pos>=l.seq() ->length())return false;
  return true;
}

void procLocs(LocList &locList,list<Location> &use,int level){
  string indent;
  for(int i=0;i<level;i++)
    indent+="  ";
  if(level<seq_count){
    for(list<Location>::iterator i=locList[level].begin();
	i!=locList[level].end();
	i++){
      list<Location> temp(use);
      assert(sanityCheck(*i));
      temp.push_back(*i);
      procLocs(locList,temp,level+1);
    }
    return;
  }
  //got a full package!
  AnchorSet *s=new AnchorSet;
  int length=mfh->hashpat->length();
  while(!use.empty()){
    //        assert(cout << " -> add "<<use.front()<< " hash: "<<
    //    	   wordToString(Window(use.front(),Hash::hashpat).hash(),debugbits) << endl);
    s->add(use.front(),length);
    use.pop_front();
  }
  assert(cout << "Matched: "<<*s << endl);
  if(opt.skip1to1){
    vector<Space>::iterator i(s->spaces.begin());
    int ref=(*i)->start;
    for(i++;i!=s->spaces.end();i++)
      if((*i)->start!=ref)
	goto HAS_OTHERS;
    //finished without finding a non-self reference
    delete s; //discard
    return; //abort
  HAS_OTHERS:; //safe ^^
  }
  //s ready!
  if(anchors->alreadyExists(*s)){
    delete s;
    //        assert(cout << "^--- Already exists...\n");
    return; //already completely included. useless
  }

  s->extend();
  if(anchors->merge(*s)){
    //    assert(cout << "^---- merged\n");
    return; //coalesced with existing blocks
  }
  //new territory!
  anchors->add(*s);
//  assert(cout << "^---- added\n");
#ifndef NDEBUG  
  writeAnchors(opt.anchor_record);
#endif
}

string platformInfo(){
  char buf[1024];
  sprintf(buf,"Platform information:\nWordsize: %u bits\n\
sizeof(word): %u bytes\n\
%s\n\
",WORDSIZE,(unsigned)sizeof(word),
	  sysinfo.toString().c_str());
  return string(buf);
}

string program_version(){
  string feature;
#ifdef MURASAKI_MYSQL
  if (feature=="") feature=" (";
  feature += "MySQL";
#endif
#ifdef MURASAKI_SQLITE3
  if (feature=="") feature=" ("; else feature += ", ";
  feature += "SQLite3";
#endif
#ifdef MURASAKI_MPI
  if (feature=="") feature=" ("; else feature += ", ";
  feature += "MPICH";
#endif
  if (feature!="") feature+=")";

  return "Murasaki version "+version_string+feature+"\n";
}

#ifdef __FreeBSD__
int getsysctl(char* name){
  int mib[4];
  size_t len = 4;
  int size;

  sysctlnametomib(name,  mib, &len);
  if (sysctl(mib, 4, &size, &len, NULL, 0) != -1)
    return size;

  return 0;
}
#endif

SystemInfo::SystemInfo()
  : wordsize(WORDSIZE),unit(1024)
{
  //platform specific mojo. 

  // for Linux:
#ifdef __linux__
  using namespace boost;
  ifstream meminfo("/proc/meminfo");
  const boost::regex memline("(\\w+):\\s+(\\d+)\\s(\\w+)");
  boost::smatch results;
  string line;

  while(getline(meminfo,line)){
    if(regex_search(line,results,memline)){
      if(results[1]=="MemTotal")
	totalMemory=lexical_cast<int>(results[2]);
      else if(results[1]=="MemFree")
	freeMemory=lexical_cast<int>(results[2]);
    }
  }
#endif

  // for FreeBSD 
#ifdef __FreeBSD__
  int pagesize = getpagesize()/unit;
  unsigned int pa, pi, pw, pc, pf;

  pa = getsysctl("vm.stats.vm.v_active_count");
  pi = getsysctl("vm.stats.vm.v_inactive_count");
  pw = getsysctl("vm.stats.vm.v_wire_count");
  pc = getsysctl("vm.stats.vm.v_cache_count");
  pf = getsysctl("vm.stats.vm.v_free_count");

  totalMemory = (pa + pi + pw + pc + pf) * pagesize;
  freeMemory = (pc + pf) * pagesize;
#endif

  // for MacOS X
#ifdef __APPLE__
  vm_statistics_data_t page_info;
  host_basic_info maxmem_info; 
  
  vm_size_t pagesize;
  mach_msg_type_number_t count;
  kern_return_t kret;
  unsigned int pf;

  pagesize = 0;
  kret = host_page_size (mach_host_self(), &pagesize);

  count = HOST_VM_INFO_COUNT;
  kret = host_statistics (mach_host_self(), HOST_VM_INFO,
			  (host_info_t)&page_info, &count);

  if (kret == KERN_SUCCESS){
    pf = page_info.free_count*pagesize;
    freeMemory = pf / unit;
  }

  count = HOST_BASIC_INFO_COUNT;
  kret = host_info (mach_host_self(), HOST_BASIC_INFO, (host_info_t)&maxmem_info, &count);
  if (kret == KERN_SUCCESS){
    totalMemory = maxmem_info.max_mem / unit;
  }
#endif

}

string SystemInfo::toString() const{
  char buf[1024];
  sprintf(buf,"Total Memory: %s\n\
Available Memory: %s (%.2f%%)\n",
	  humanMemory(totalMemory*unit).c_str(),
	  humanMemory(freeMemory*unit).c_str(),
	  ((double)freeMemory/(double)totalMemory*100.0));
  return string(buf);
}

ostream& operator<<(ostream& of,const SystemInfo& a){
  return of << a.toString();
}

ostream& outputRegions(ostream &of){
  for(unsigned i=0;i<seqs.size();i++){
    BitSequence *a=seqs[i]->fwd;
  start: of << "Seq "<<i<<" "<< (seqs[i]->name)<<" strand "<<(a==seqs[i]->fwd ? "fwd":"rev")<<endl;
    for(unsigned j=0;j<a->matchRegions.size();j++)
      of << " Region from: "<<a->matchRegions[j].first
	   <<" to "<<a->matchRegions[j].second<<endl;
    if(a==seqs[i]->fwd){
      a=seqs[i]->rev;
      goto start;
    }
  }
  return of;
}

void hashStatusCheck(int tick,int pos){
#ifdef NDEBUG
  return; //this function is really slow actually. rely on tee instead
#endif
  ofstream of(opt.hashStatus_record.c_str());
  of << "Done "<<tick<<"%"<<" ("<<pos<<")"<<endl;
  of << "Anchors: "<<anchors->count()<<endl;
  //  of << SystemInfo();
}

//for calling from gdb
void verboseOn(){
  verbose=1;
}

void verboseOff(){
  verbose=0;
}

bool convertPattern(string &pat){
  int weight,length;
  if(sscanf(pattern_src.c_str(),"[%d:%d]",&weight,&length)!=2 &&
     sscanf(pattern_src.c_str(),"[%d,%d]",&weight,&length)!=2)
    return false;
  //make a random pattern
  string cmd("./randpat.pl");
  vector<string> argv(1);
  argv.push_back(string("-c"));
  argv.push_back(dstring(weight));
  argv.push_back(dstring(length));
  cout << "Generating random pattern weight "<<weight<<" and length "<<length<<endl;
  redi::ipstream inf(cmd,argv);
  inf >> pat;
  return true;
}

